/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2015                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/
#include "FontWidget.h"

#include <QLabel>
#include <QVBoxLayout>
#include <QDialogButtonBox>
#include <QPushButton>
#include <QTreeView>
#include <QDebug>

FontWidget::FontWidget( QWidget*       parent,
                        const QString& family,
                        int            pointSize ) : QDialog( parent )
{
    setWindowTitle( "Font settings" );

    QVBoxLayout* layout = new QVBoxLayout();
    setLayout( layout );

    QLabel* fontLabel = new QLabel( this );
    fontLabel->setText( "Font: " );
    layout->addWidget( fontLabel );

    fontCombo = new QFontComboBox( this );
    layout->addWidget( fontCombo );
    if ( family != USE_DEFAULT )
    {
        fontCombo->setCurrentFont( QFont( family ) );
        fontCombo->insertItem( 0, "Use Qt default", QTreeView().font() );
    }
    else
    {
        fontCombo->insertItem( 0, "Use Qt default", QTreeView().font() );
        fontCombo->setCurrentIndex( 0 );
    }

    QLabel* sizeLabel = new QLabel( this );
    sizeLabel->setText( "Size [pt]: " );
    layout->addWidget( sizeLabel );

    sizeCombo = new  QComboBox();
    sizeCombo->addItem( "Use Qt default" );
    for ( int i = 6; i <= 20; i++ )
    {
        sizeCombo->addItem( QString::number( i ) );
        if ( i == pointSize )
        {
            sizeCombo->setCurrentIndex( sizeCombo->count() - 1 );
        }
    }
    if ( pointSize == -1 )
    {
        sizeCombo->setCurrentIndex( 0 );
    }
    layout->addWidget( sizeCombo );

    QDialogButtonBox* buttonBox = new QDialogButtonBox();
    buttonBox->addButton( QDialogButtonBox::Ok );
    QPushButton* applyButton = buttonBox->addButton( QDialogButtonBox::Apply );
    buttonBox->addButton( QDialogButtonBox::Cancel );

    connect( buttonBox, SIGNAL( accepted() ), this, SLOT( onOk() ) );
    connect( applyButton, SIGNAL( clicked() ),  this, SLOT( onApply() ) );
    connect( buttonBox, SIGNAL( rejected() ), this, SLOT( onCancel() ) );

    layout->addWidget( buttonBox );

    setWhatsThis( "Opens a dialog to specify the font, the font size (in pt), and the line spacing for the tree displays.  The \"Ok\" button applies the settings to the display and closes the dialog, the \"Apply\" button applies the settings to the display, and \"Cancel\" cancels all changes since the dialog was opened (even if \"Apply\" was pressed in between) and closes the dialog." );

    familyIndex = fontCombo->currentIndex();
    sizeIndex   = sizeCombo->currentIndex();
}

QFont
FontWidget::getFont()
{
    QFont font = ( fontCombo->currentIndex() > 0 ) ? fontCombo->currentFont() : QTreeView().font();
    int   size = ( sizeCombo->currentIndex() > 0 ) ? sizeCombo->currentText().toInt() : QTreeView().font().pointSize();
    font.setPointSize( size );
    return font;
}

QString
FontWidget::getFontFamily()
{
    return ( fontCombo->currentIndex() > 0 ) ? fontCombo->currentFont().family() : USE_DEFAULT;
}

int
FontWidget::getFontSize()
{
    return ( sizeCombo->currentIndex() > 0 ) ? sizeCombo->currentText().toInt() : -1;
}

void
FontWidget::onOk()
{
    onApply();
    accept();
}

void
FontWidget::onApply()
{
    emit apply( this );
}

void
FontWidget::onCancel()
{
    fontCombo->setCurrentIndex( familyIndex );
    sizeCombo->setCurrentIndex( sizeIndex );
    onApply();
    reject();
}
