#ifndef TABWIDGETCONTAINER_H
#define TABWIDGETCONTAINER_H

#include <QSplitter>
#include "Constants.h"
#include "Cube.h"
#include "TabWidget.h"

class Tree;
class TreeItem;

/** This class manages the 3 TabWidgets of type METRICTAB, CALLTAB and SYSTEMTAB.
 *  - it handles the tree selections, as this may have effect on other trees
 *
 */
class TabManager : public QSplitter
{
    Q_OBJECT
public:
    /**
     * creates an empty container for the three tabs with their trees.
     */
    TabManager( QWidget* parent = 0 );

    /** To be called after the settings have been processed. Ensures that at least one item is selected in
     *  all trees, connects signals and calculates all tree values */
    void
    initialize();

    /** recalculates all tree values */
    void
    reinit();

    /**
     * Initializes the container with the given cube data.
     * Creates the four trees and its views and initializes them.
     */
    void
    cubeOpened( cube::Cube* cube );
    void
    cubeClosed();

    QList<DisplayType>
    getOrder() const;
    void
    setOrder( const QList<DisplayType>& value );

    TabWidget*
    getTab( DisplayType tabType );

    Tree*
    getTree( TreeType type );

    TreeView*
    getView( TreeType type );

    // SettingsHandler interface
    void
    loadGlobalSettings( QSettings& )
    {
    };
    void
    saveGlobalSettings( QSettings& )
    {
    };
    void
    loadExperimentSettings( QSettings& );
    void
    saveExperimentSettings( QSettings& );
    QString
    settingName();

    // end SettingsHandler interface

    void
    setTreeFont( const QFont& font,
                 int          spacing );

    QList<TabWidget*>
    getTabWidgets();

    /** returns the tree in the active (visible) tab */
    Tree*
    getActiveTree( DisplayType tabType );

    DisplayType
    getDisplayType( TreeType treeType );

public slots:
    /**
     * redraws all tree items with their changed properties, e.g. color, precision
     */
    void
    updateTreeItemProperties();

private slots:
    void
    treeItemSelected( Tree* tree );
    void
    treeItemExpanded( TreeItem* item,
                      bool      expanded );
    void
    recalculateTree( Tree* tree );

private:
    QList<Tree*>                 trees;       // the 4 trees: METRICTREE, CALLTREE, CALLFLAT, SYSTEMTREE
    QList<TreeView*>             views;       // one view for each tree
    QList<TabWidget*>            tabWidgets;  // the three TabWidget: METRICTAB, CALLTAB, SYSTEMTAB
    QList<QAbstractItemModel*>   proxyModels; // delete all models after cube is closed
    QMap<TreeType, DisplayType>  typeHash;
    QList<DisplayType>           order;

    // --------------------------------------------------------------

    void
    createTreeAndView( TreeType    type,
                       cube::Cube* cube );
    void
    getNeighborTrees( QList<Tree*>& left,
                      QList<Tree*>& right,
                      Tree*         current );
};

#endif // TABWIDGETCONTAINER_H
